import argparse
from models.model_builder import MODEL_REGISTRY

def arg_parser():
    parser = argparse.ArgumentParser(description='PyTorch Action recognition Training')

    # model definition
    parser.add_argument('--backbone_net', default='resnet18', type=str, help='backbone network',
                        choices=list(MODEL_REGISTRY._obj_map.keys()))
    parser.add_argument('--dropout', default=0.5, type=float,
                        help='dropout ratio before the final layer')
    
    # fixmatch setting
    parser.add_argument('--hard_label', type=bool, default=True)
    parser.add_argument('--T', type=float, default=0.5)
    parser.add_argument('--p_cutoff', type=float, default=0.8)
    parser.add_argument('-u', '--ulb_loss_ratio', type=float, default=0.0)
    parser.add_argument('--uratio', type=int, default=1)
    parser.add_argument('--nlabels', type=int, default=None)
    parser.add_argument('--fixed_order', action='store_true')
    parser.add_argument('--unlabeled_folder', type=str, default=None, help = "unlabeled dataset")
    
    # over confidence
    parser.add_argument('--T_scaling', action='store_true')
    
    # RL part
    parser.add_argument('--pseudo_label', type=str, default=None, help = "Add pseudo label from RL")
    parser.add_argument('--traj_power', type=int, default=4, help = "Power rate for weight")
    
    # algorithms
    parser.add_argument('--alg', type=str, default="fixmatch", help = "unsupervised learning algorithm (fixmatch or paws)")
    
    
    # training setting
    parser.add_argument('--seed', type = int, default= None, help='random seed for replication')
    parser.add_argument('--gpu', help='comma separated list of GPU(s) to use.')
    parser.add_argument('--gpu_id', help='comma separated list of GPU(s) to use.', default=None)
    parser.add_argument('--disable_cudnn_benchmark', dest='cudnn_benchmark', action='store_false',
                        help='Disable cudnn to search the best mode (avoid OOM)')
    parser.add_argument('-b', '--batch-size', default=256, type=int,
                        metavar='N', help='mini-batch size (default: 256)')
    parser.add_argument('--lr', '--learning-rate', default=0.01, type=float,
                        metavar='LR', help='base learning rate')
    parser.add_argument('--warmup_lr', default=None, type=float,
                        metavar='LR', help='warmup learning rate')
    parser.add_argument('--lr_scheduler', default='cosine', type=str,
                        help='learning rate scheduler',
                        choices=['step', 'multisteps', 'cosine', 'plateau', 'warmupcosine'])
    parser.add_argument('--lr_steps', default=[15, 30, 45], type=float, nargs="+",
                        metavar='LRSteps', help='[step]: use a single value: the periodto decay '
                                                'learning rate by 10. '
                                                '[multisteps] epochs to decay learning rate by 10')
    parser.add_argument('--momentum', default=0.9, type=float, metavar='M', help='momentum')
    parser.add_argument('--nesterov', action='store_true',
                        help='enable nesterov momentum optimizer')
    parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float,
                        metavar='W', help='weight decay (default: 5e-4)')
    parser.add_argument('--epochs', default=50, type=int, metavar='N',
                        help='number of total epochs to run')
    parser.add_argument('--warmup_epochs', default=5, type=int, metavar='N',
                        help='number of total epochs for warmup')
    parser.add_argument('--resume', default='', type=str, metavar='PATH',
                        help='path to latest checkpoint (default: none)')
    parser.add_argument('--pretrained', dest='pretrained', type=str, metavar='PATH',
                        help='use pre-trained model')
    parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                        help='manual epoch number (useful on restarts)')
    parser.add_argument('--clip_gradient', '--cg', default=None, type=float,
                        help='clip the total norm of gradient before update parameter')
    parser.add_argument('--label_smoothing', default=0.0, type=float, metavar='SMOOTHING',
                        help='label_smoothing against the cross entropy')
    parser.add_argument('--no_imagenet_pretrained', dest='imagenet_pretrained',
                        action='store_false',
                        help='disable to load imagenet pretrained model')
    parser.add_argument('--transfer', action='store_true',
                        help='perform transfer learning, remove weights in the fc '
                             'layer or the original model.')
    parser.add_argument('--auto_resume', action='store_true', help='if the log folder includes a checkpoint, automatically resume')
    parser.add_argument('--lazy_eval', action='store_true', help="evaluate every 10 epochs and last 10 percentage of epochs")
    parser.add_argument('--precise_bn', action='store_true', help="use precise bn to get better stats of bn layer")
    parser.add_argument('--patch_eval', type = int, nargs='+', default= None, help = 'select a specific patch to evaluate')
    parser.add_argument('--patch_train', type = int, nargs='+', default= None, help = 'select a specific patch to train')
    parser.add_argument('--opt', type = str, default= "SGD", help = 'Optimizer')
    
    # handling imbalanced data
    parser.add_argument('--weighted_data', type=str, default= None, help='path to the file that handls imbanlanc data. If `by_class` is given then balance based on class')
    parser.add_argument('--eff_size', type=float, default = 1, help='ratio between true label and unlabel')
    

    # data-related
    parser.add_argument('-c', '--config', type=str, help='YAML setting for the dataset')
    parser.add_argument('--dataset', type = str, default=None, help = "name for the dataset")
    parser.add_argument('--datadir', default=None, type=str, help='Path to the dataset')
    parser.add_argument('-j', '--workers', default=18, type=int, metavar='N',
                        help='number of data loading workers (default: 4)')
    parser.add_argument('--input_size', default=224, type=int, metavar='N', help='input image size')
    parser.add_argument('--mean', type=float, nargs="+",
                        metavar='MEAN', help='mean, dimension should be 3 for RGB, 1 for flow')
    parser.add_argument('--std', type=float, nargs="+",
                        metavar='STD', help='std, dimension should be 3 for RGB, 1 for flow')
    parser.add_argument('--skip_normalization', action='store_true',
                        help='skip mean and std normalization, default use imagenet`s mean and std.')
    parser.add_argument('--half_data', action='store_true', help='whether to use half data for patch')
    parser.add_argument('--val_folder', type=str, default=None, help = "validation dataset")

    # logging
    parser.add_argument('--logdir', default='./log', type=str, help='log path')
    parser.add_argument('--abs_log_dir', default=None, type=str, help='abs log path. Will cover all the other dirs')
    parser.add_argument('--prefix', default='', type=str, help='prefix in the logdir')
    parser.add_argument('--print-freq', default=100, type=int,
                        help='frequency to print the log during the training')
    parser.add_argument('--show_model', action='store_true', help='show model summary')
    parser.add_argument('--group_dir', default=None, type = str, help='used to generate some summaries by grouping experiments')

    # for testing and validation
    parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                        help='evaluate model on validation set')
    parser.add_argument('--bayesian', action='store_true',
                        help='use Bayesian network to calibrate the overconfidence')
    parser.add_argument('--cpu', action='store_true', help='using cpu only')

    # for distributed learning
    parser.add_argument('--sync-bn', action='store_true',
                        help='sync BN across GPUs')
    parser.add_argument('--world-size', default=1, type=int,
                        help='number of nodes for distributed training')
    parser.add_argument('--rank', default=0, type=int,
                        help='node rank for distributed training')
    parser.add_argument('--hostfile', default='', type=str,
                        help='hostfile distributed learning')
    parser.add_argument('--dist-url', default='tcp://127.0.0.1:23456', type=str,
                        help='url used to set up distributed training')
    parser.add_argument('--dist-backend', default='nccl', type=str,
                        help='distributed backend')
    parser.add_argument('--multiprocessing-distributed', '--ddp', action='store_true',
                        help='Use multi-processing distributed training to launch '
                             'N processes per node, which has N GPUs. This is the '
                             'fastest way to use PyTorch for either single node or '
                             'multi node data parallel training')
    return parser


if __name__ == '__main__':
    parser = arg_parser()
    args = parser.parse_args()

    print(args)
