# evaluate 
# a pesudo training framework
from ast import arg
import os
import argparse
from datetime import datetime
from os.path import join
import re
import numpy as np

# evaluate runs in terms of RL rewards



root_path = os.getcwd()

parser = argparse.ArgumentParser(description='CryoEM_Pseudo_Training')
parser.add_argument('--nlabels', type=int, default=125)
parser.add_argument('--seed', type=int, default=3)
parser.add_argument('--dataset', type=str, default="Y3_2Cat")
parser.add_argument('--pretrained', type=str, default=None)
parser.add_argument('--fixed_order', action="store_true")
parser.add_argument('--repeat', type=int, default=5)
parser.add_argument('--gpu_id', type = int, help='comma separated list of GPU(s) to use.', default=0)
parser.add_argument('--sub_dir', nargs="*", type = str, default= None, help='evaluate a specific subdir')

# the following are unused
parser.add_argument('--ctf_thred', type=int, default=6)
parser.add_argument('--nplabels', type=int, default=100)
parser.add_argument('--fine_tune', action='store_true')
parser.add_argument('-e', type=int, default=100)
parser.add_argument('-b', type=int, default=16)
parser.add_argument('--num_rounds', type = int, default=10)
parser.add_argument('--lr', type = float, default=0.001)
parser.add_argument('--use_one_hot', action="store_true")
parser.add_argument('--bayesian', action="store_true", help = "use bayesian calibration")
parser.add_argument('--supervised', action="store_true")
parser.add_argument('--use_clf', action="store_true")
parser.add_argument('--traj_num', type=int, default= 5)
parser.add_argument('--duration', type=int, default= 120)
parser.add_argument('--patch', action="store_true")
parser.add_argument('--confidence', type = float, default=0, help="keep samples max(p) > confidence")
parser.add_argument('--eff_size', type = int, default=1, help="ratio between labeled and pesudo labeled")
parser.add_argument('--traj_power', type = int, default=4, help="power for trajectory weight")

global args
args = parser.parse_args()

if not os.path.exists(args.pretrained):
    args.model_path = os.path.join(root_path, "CryoEM_Img", "log/timeBased", args.pretrained)

os.chdir("CryoEM_Img")

if args.use_one_hot:
    use_one_hot = "--use_one_hot"
else:
    use_one_hot = ""
if args.fixed_order:
    fixed_order = "--fixed_order"
else:
    fixed_order = ""


    

if args.dataset == "Y1_3Cat":
    args.config_file = "Y1_3Lab_48.yaml"
    args.RL_dataset = "Y1_3Cat"
    args.thred = "4 8"
    args.ctf_file = "CryoEM_Config/Y1_CTF_by_hl.csv"
elif args.dataset == "Y1_2Cat":
    if args.patch:
        args.config_file = "Y1_2Cat_6_Patch.yaml"
    else:
        args.config_file = "Y1_2Cat_6.yaml"
    args.RL_dataset = "Y1_2Cat"
    args.thred = "6"
    args.ctf_file = "CryoEM_Config/Y1_CTF_by_hl.csv"
elif args.dataset == "Y3_2Cat":
    if args.patch:
        args.config_file = "Y3_2Cat_6_Patch.yaml"
    else:
        args.config_file = "Y3_SL.yaml"
    args.RL_dataset = "Y3_2Cat"
    args.thred = "6"
    args.ctf_file = "CryoEM_Config/Y3_CTF_by_hl.csv"
else:
    raise NotImplementedError

start = 0
dirs = [name for name in os.listdir(args.model_path) if os.path.isdir(os.path.join(args.model_path, name))]
dirs.sort()
dirs = [""] + dirs

def get_img_acc(output):
    idf = "Validation Acc: "
    n = len(idf)
    o = re.search(idf, output)
    return float(output[o.start()+n:])

def get_RL_acc(output):
    idf = "Final reward: "
    n = len(idf)
    o = re.search(idf, output)
    e = re.search("\\+", output)
    return float(output[o.start()+n: e.start()-1])

if args.sub_dir is not None:
    dirs = args.sub_dir
for sub_dir in dirs:
    rwds = []
    os.chdir(root_path)
    os.chdir("CryoEM_Img")
    if os.path.exists(join(args.model_path, sub_dir, "evaluate_trajectory_5_runs.txt")):
        os.system("rm %s"%(join(args.model_path, sub_dir, "evaluate_trajectory_5_runs.txt")))
    for i in range(2*args.traj_num, 3*args.traj_num):
        os.chdir("../CryoRL-pytorch")
        # evaluate RL on training and generate trajectory
        os.system("python train.py --prediction_file %s --dataset %s --lr 0.001 --training-num 10 --test-num 10 --step-per-epoch 500 --ctf-thresh %d  --train-prediction --test-prediction --hard_step 10000 --duration 480 --eval --print-trajectory %s --seed %d > /dev/null" % (join(args.model_path, sub_dir), args.RL_dataset, args.ctf_thred, use_one_hot, i))
        trj = np.loadtxt(join(args.model_path, sub_dir, "evaluate_trajectory.txt"), delimiter=' ', dtype=str)
        if args.nplabels > trj.shape[0]/2:
            args.nplabels = int(trj.shape[0]/2)
        trj_pos = trj[:args.nplabels, :]
        trj_pos = np.concatenate(((np.zeros(args.nplabels).reshape(-1, 1)+1).astype(int).astype(str), trj_pos), axis = 1)
        trj_neg = trj[-args.nplabels:, :]
        trj_neg = np.concatenate(((np.zeros(args.nplabels).reshape(-1, 1)).astype(int).astype(str), trj_neg), axis = 1)
        trj = np.concatenate([trj_pos, trj_neg])
        np.savetxt(join(args.model_path, sub_dir, "evaluate_trajectory.txt"), trj, fmt = '%s')
        os.system('cat %s >> %s' % (join(args.model_path, sub_dir, "evaluate_trajectory.txt"), join(args.model_path, sub_dir, "evaluate_trajectory_5_runs.txt")))
    os.chdir("../CryoEM_Img")
    os.system("python tools/generate_pesudo.py --file %s --source from_traj --out %s" % (join(args.model_path, sub_dir, "evaluate_trajectory_5_runs.txt"), join(args.model_path, sub_dir, "pesudo_labels_appx.csv")))
    os.chdir("../CryoEM_Img")
    dat = np.loadtxt(join(args.model_path, sub_dir, "pesudo_labels_appx.csv"), delimiter = ',', dtype = str)
    avg_conf = np.sum(dat[:, 2].astype(float)) / dat.shape[0]
    with open(join(args.model_path, 'apprx_RL.txt'), 'a') as f:
        print(f"{sub_dir}, {avg_conf}", file = f)
    