# a pesudo training framework
from ast import arg
import os
import sys
import csv
import argparse
from datetime import datetime
from os.path import join
import re
import numpy as np
import logging
from sklearn import metrics

root_path = os.getcwd()

    
os.chdir(root_path)
parser = argparse.ArgumentParser(description='CryoEM_Pseudo_Training')
parser.add_argument('--nlabels', type=int, default=125)
parser.add_argument('--seed', type=int, default=3)
parser.add_argument('--dataset', type=str, default="Y3_2Cat")
parser.add_argument('--pretrained', type=str, default=None)
parser.add_argument('--fixed_order', action="store_true")
parser.add_argument('--gpu_id', type = int, help='comma separated list of GPU(s) to use.', default=0)

# the following are unused
parser.add_argument('--ctf_thred', type=int, default=6)
parser.add_argument('--nplabels', type=int, default=100)
parser.add_argument('--fine_tune', action='store_true')
parser.add_argument('-e', type=int, default=100)
parser.add_argument('-b', type=int, default=16)
parser.add_argument('--num_rounds', type = int, default=10)
parser.add_argument('--lr', type = float, default=0.001)
parser.add_argument('--use_one_hot', action="store_true")
parser.add_argument('--bayesian', action="store_true", help = "use bayesian calibration")
parser.add_argument('--supervised', action="store_true")
parser.add_argument('--use_clf', action="store_true")
parser.add_argument('--traj_num', type=int, default= 5)
parser.add_argument('--duration', type=int, default= 120)
parser.add_argument('--patch', action="store_true")
parser.add_argument('--confidence', type = float, default=0, help="keep samples max(p) > confidence")
parser.add_argument('--eff_size', type = int, default=1, help="ratio between labeled and pesudo labeled")
parser.add_argument('--traj_power', type = int, default=4, help="power for trajectory weight")
global args
args = parser.parse_args()

os.chdir("CryoEM_Img")

if args.fixed_order:
    fixed_order = "--fixed_order"
else:
    fixed_order = ""

if args.dataset == "Y1_3Cat":
    args.config_file = "Y1_3Lab_48.yaml"
    args.RL_dataset = "Y1_3Cat"
    args.thred = "4 8"
    args.ctf_file = "CryoEM_Config/Y1_CTF_by_hl.csv"
elif args.dataset == "Y1_2Cat":
    if args.patch:
        args.config_file = "Y1_2Cat_6_Patch.yaml"
    else:
        args.config_file = "Y1_2Cat_6.yaml"
    args.RL_dataset = "Y1_2Cat"
    args.thred = "6"
    args.ctf_file = "CryoEM_Config/Y1_CTF_by_hl.csv"
elif args.dataset == "Y3_2Cat":
    if args.patch:
        args.config_file = "Y3_2Cat_6_Patch.yaml"
    else:
        args.config_file = "Y3_SL.yaml"
    args.RL_dataset = "Y3_2Cat"
    args.thred = "6"
    args.ctf_file = "CryoEM_Config/Y3_CTF_by_hl.csv"
else:
    raise NotImplementedError

args.model_path = join(root_path, "CryoEM_Img", "log", "timeBased", args.pretrained)
start = 0
dirs = [name for name in os.listdir(args.model_path) if os.path.isdir(os.path.join(args.model_path, name))]
dirs.sort()
dirs = [""] + dirs

def get_metrics(vec):
    fpr, tpr, thresholds = metrics.roc_curve(vec['true'], vec['score'], pos_label=0)
    auc = metrics.auc(fpr, tpr)
    prec_score = metrics.precision_score(vec['true'], vec['pred'])
    acc = metrics.accuracy_score(vec['true'], vec['pred'])
    f1 = metrics.f1_score(vec['true'], vec['pred'])
    rcl = metrics.recall_score(vec['true'], vec['pred'])
    blc_acc = metrics.balanced_accuracy_score(vec['true'], vec['pred'])
    # print(metrics.classification_report(vec['true'], vec['pred']))
    return (acc, blc_acc, auc, prec_score, f1, rcl)

f = open(join(args.model_path, 'clf_summary.csv'), 'w')
writter = csv.writer(f, delimiter=',')
writter.writerow(['round', 'acc', 'blc_acc', 'auc', 'prec', 'f1', 'rcl'])
train_res = []
val_res = []
for sub_dir in dirs:
    rwds = []
    os.chdir(root_path)
    os.chdir("CryoEM_Img")
    
    # regenerate train_by_hl and val_by_hl files
    os.system("python train.py --config configs/%s --seed %d --nlabels %d -e --pretrained %s --backbone_net resnet18 --gpu_id %d %s" % (args.config_file, args.seed, args.nlabels, join(args.model_path, sub_dir, "model_best.pth.tar"), args.gpu_id, fixed_order))
    
    # load by_hl files
    train_hl = np.loadtxt(join(args.model_path, sub_dir, "train_by_hl.txt"), dtype = 'str')
    val_hl = np.loadtxt(join(args.model_path, sub_dir, "val_by_hl.txt"), dtype = 'str')
    
    convert = lambda x: int(x == "low")
    
    train_vec = {'pred': np.array(list(map(convert, train_hl[:, -1]))),
                 'true': np.array(list(map(convert, train_hl[:, -2]))), 
                 'score': np.array(list(train_hl[:, -4])).astype('float')}
    val_vec = {'pred': np.array(list(map(convert, val_hl[:, -1]))),
               'true': np.array(list(map(convert, val_hl[:, -2]))), 
               'score': np.array(list(val_hl[:, -4])).astype('float')}
    res = [sub_dir+'train']
    res += list(get_metrics(train_vec))
    # writter.writerow(res)
    train_res.append(res)
    res = [sub_dir+'val']
    res += list(get_metrics(val_vec))
    # writter.writerow(res)
    val_res.append(res)
writter.writerows(train_res)
writter.writerows(val_res)
f.close()
os.system('cat %s'%(join(args.model_path, 'clf_summary.csv')))
    