from dataclasses import replace

from torchvision.datasets import ImageFolder
import numpy as np
import os
from os.path import join
import shutil
import re
import argparse

def combine_dataset(datasets: list):
    samples = []
    targets = []
    for dset in datasets:
        samples += dset.samples
        targets += dset.targets
    datasets[0].samples = samples
    datasets[0].targets = targets
    # datasets[0].regenerate_data()
    return datasets[0]

parser = argparse.ArgumentParser(description='Generate fix ordering')
parser.add_argument('--root_dir', type=str, default=None)
parser.add_argument('--p', type=float, nargs="*", default=[0.05, 0.1, 0.2, 1])
args = parser.parse_args("--root_dir ../data/Y3_2Cat6/train".split())

output_file = "order.txt"
dataset = ImageFolder(args.root_dir)
samples = dataset.samples
num = len(samples)
path_list = []
target_list = []
for path, target in samples:
    path_list.append(path)
    target_list.append(target)
high_idx = [i for i in range(num) if target_list[i] == 0]
low_idx = [i for i in range(num) if target_list[i] == 1]

ratio_high = len(high_idx) / (num)
ratio_low = 1 - ratio_high

# generate
p_cum = 0 
high_cum = 0
low_cum = 0
idx_list = []
# shuffle
high_idx = np.random.choice(high_idx, len(high_idx), replace = False)
low_idx = np.random.choice(low_idx, len(low_idx), replace = False)
for p in args.p:
    add = (num * (p - p_cum))
    add_high = int(add * ratio_high)
    add_low = int(add * ratio_low)
    idx_list += list(high_idx[high_cum:(high_cum + add_high)])
    idx_list += list(low_idx[low_cum:(low_cum + add_low)])
    high_cum += add_high
    low_cum += add_low
    p_cum = p

np.savetxt(join(args.root_dir, output_file), idx_list, fmt = "%d", delimiter = ",")
    