# a pesudo training framework
from ast import arg
import os
import csv
import sys
import argparse
from datetime import datetime
from os.path import join
import re
import numpy as np
import logging
import torch
from collections import Counter
from torchnet import meter

def find_majority(votes):
    vote_count = Counter(votes)
    top_two = vote_count.most_common(2)
    if len(top_two)>1 and top_two[0][1] == top_two[1][1]:
        # It is a tie
        return 0
    return top_two[0][0]

def get_avg(confusion):
    num = np.sum(confusion.conf)
    return np.sum(np.diag(confusion.value()) * np.sum(confusion.conf, axis = 1)) / num, num

parser = argparse.ArgumentParser(description='Generate_Pesudo_Data')
parser.add_argument('--file', type=str)
parser.add_argument('--source', type=str, default="from_traj", choices=["from_traj", "from_clf"])
parser.add_argument('--out', type=str, default=None)
parser.add_argument('--traj_power', type=int, default=4)
parser.add_argument('--eval', action='store_true')

args = parser.parse_args()
if args.out is None:
    args.out = join(os.path.dirname(args.file), "pesudo_labels.csv")

if args.source == "from_traj":
    file = open(args.file, 'r')
    traj = csv.reader(file, delimiter =' ')
    traj = [row for row in traj]
    file.close()
    names = [row[4] for row in traj]
    plabels = [int(row[0]) for row in traj]
    tlabels = [int(float(row[9][1:]) < 6) for row in traj]
    # true_labels = [int(float(row[9][1:]) > 6) for row in traj]
    # get idx
    cnt_dict = {}
    p_dict = {} # proportion of majority
    lab_dict = {}
    tlab_dict = {}
    for i in range(len(names)):
        if names[i] in cnt_dict.keys():
            cnt_dict[names[i]]+=1
            lab_dict[names[i]].append(plabels[i])
        else:
            cnt_dict[names[i]]=1
            tlab_dict[names[i]] = tlabels[i]
            lab_dict[names[i]] = []
            lab_dict[names[i]].append(plabels[i])
    for i in range(len(names)):
        if type(lab_dict[names[i]]) == list:
            p_dict[names[i]] = sum(lab_dict[names[i]]) / len(lab_dict[names[i]])
            if p_dict[names[i]] < 0.5:
                p_dict[names[i]] = 1 - p_dict[names[i]]
            lab_dict[names[i]] = find_majority(lab_dict[names[i]])
    if args.eval:
        confusion = meter.ConfusionMeter(2, normalized=True)
        confusion.add(torch.Tensor(list(lab_dict.values())), torch.Tensor(list(tlab_dict.values())))
        print("---confusion matrix")
        print(confusion.value())
        print(confusion.conf)
        print("Average: ", (get_avg(confusion)))
    else:
        file = open(args.out, 'w')
        for key in cnt_dict.keys():
            writer = csv.writer(file)
            writer.writerow([key, lab_dict[key], p_dict[key]**args.traj_power])
    file.close()
elif args.source == "from_clf":
    file = open(args.file, 'r')
    traj = csv.reader(file, delimiter =' ')
    traj = [row for row in traj]
    file.close()
    names = [row[0] for row in traj]
    lab_dict = {'low': 1, 'high': 0}
    plabels = [lab_dict[(row[-1])] for row in traj]
    tlabels = [lab_dict[(row[-2])] for row in traj]
    weight = [max(float(row[1]), float(row[2])) for row in traj]
    if args.eval:
        plabels = torch.Tensor(plabels)
        tlabels = torch.Tensor(tlabels)
        confusion = meter.ConfusionMeter(2, normalized=True)
        confusion.add(plabels, tlabels)
        print("---confusion matrix")
        print(confusion.value())
        print(confusion.conf)
        print("Average: ", (get_avg(confusion)))
    else:
        file = open(args.out, 'w')
        for i in range(len(names)):
            writer = csv.writer(file)
            writer.writerow([names[i], plabels[i], weight[i]])
        file.close()