from http.client import NotConnected
import numpy as np
import math
import re
import json
import os
import csv
from os.path import join
from torchvision.datasets import ImageFolder
from utils.fixmatch_utils import fixMatchDataset

from collections import Counter

def find_majority(votes):
    vote_count = Counter(votes)
    top_two = vote_count.most_common(2)
    if len(top_two)>1 and top_two[0][1] == top_two[1][1]:
        # It is a tie
        return 0
    return top_two[0][0]

fixed_path_order  = range(25)
nSplit = 25

def load_pseudo(path):
    file = open(path, 'r')
    traj = csv.reader(file, delimiter =',')
    traj = [row for row in traj]
    file.close()
    names = [row[0] + "_crop.png" for row in traj]
    plabels = [int(row[1]) for row in traj]
    weight = [float(row[2]) for row in traj]
    # np.sum(np.array(tlabels) == np.array(plabels)) / len(plabels)
    # print(names)
    return names, plabels, weight

def load_pseudo_old(path):
    file = open(path, 'r')
    traj = csv.reader(file, delimiter =' ')
    traj = [row for row in traj]
    file.close()
    names = [row[4] + "_crop.png" for row in traj]
    plabels = [int(row[0]) for row in traj]
    tlabels = [int(float(row[9][1:]) > 6) for row in traj]
    # np.sum(np.array(tlabels) == np.array(plabels)) / len(plabels)
    # print(names)
    return names, plabels

def add_pseudo_data(full_dataset: ImageFolder, data2add: ImageFolder, names: list, plabels: list, weight: list):
    """Add pseudo data to a dataset

    Args:
        dataset (ImageFolder): An torchvision dataset
        names (list): names of the images to be added
    """
    # get idx
    cnt_dict = {}
    p_dict = {} # proportion of majority
    lab_dict = {}
    weight_dict = {}
    for i in range(len(names)):
        if names[i] in cnt_dict.keys():
            cnt_dict[names[i]]+=1
            lab_dict[names[i]].append(plabels[i])
            weight_dict[names[i]].append(weight[i])
        else:
            cnt_dict[names[i]]=1
            lab_dict[names[i]] = []
            lab_dict[names[i]].append(plabels[i])
            weight_dict[names[i]] = []
            weight_dict[names[i]].append(weight[i])
    # remove duplicates
    for i in range(len(names)):
        if type(lab_dict[names[i]]) == list:
            p_dict[names[i]] = sum(lab_dict[names[i]]) / len(lab_dict[names[i]])
            if p_dict[names[i]] < 0.5:
                p_dict[names[i]] = 1 - p_dict[names[i]]
            lab_dict[names[i]] = find_majority(lab_dict[names[i]])
            weight_dict[names[i]] = sum(weight_dict[names[i]]) / len(weight_dict[names[i]])
    full_path = [os.path.basename(sample[0]) for sample in full_dataset.samples]
    add_path = [os.path.basename(sample[0]) for sample in data2add.samples]
    
    names = list(set(names) - set(add_path)) # find those that are not in to add list
    # print("full path len:", full_path[:10])
    lb_idx = list(filter(lambda i: full_path[i] in names, range(len(full_path))))
    # print(lb_idx)
    num = len(lb_idx)
    counts = []
    props = []
    weights = []
    for i in range(num):
        tmp, true_lab = full_dataset.samples[lb_idx[i]]
        file_name = full_path[lb_idx[i]]
        data2add.samples.append((tmp, lab_dict[file_name]))
        data2add.targets.append(lab_dict[file_name]) # mark as low
        # data2add.samples.append((tmp, true_lab))
        # data2add.targets.append(true_lab) # mark as low
        data2add.imgs.append(full_dataset.imgs[lb_idx[i]])
        counts.append(cnt_dict[file_name])
        props.append(p_dict[file_name])
        weights.append(weight_dict[file_name])
    return data2add, counts, props, weights
    
    
    
def combine_dataset(datasets: list):
    samples = []
    targets = []
    for dset in datasets:
        samples += dset.samples
        targets += dset.targets
    datasets[0].samples = samples
    datasets[0].targets = targets
    datasets[0].regenerate_data()
    return datasets[0]

def dump_json(obj, path):

    # create json object from dictionary
    json_file = json.dumps(obj)
    f = open(path,"w")
    f.write(json_file)
    f.close()

def split_ssl_data(target, num_labels, num_classes = 2, determined = False):
    """
    data & target is splitted into labeled and unlabeld data.
    
    Args
        index: If np.array of index is given, select the data[index], target[index] as labeled samples.
        include_lb_to_ulb: If True, labeled data is also included in unlabeld data
    """
    # assert num_labels % num_classes == 0
    samples_per_class = num_labels * np.unique(target, return_counts = True)[1] / len(target)
    num_classes = len(np.unique(target, return_counts = True)[1])
    # print(samples_per_class)
    lb_idx = []
    for c in range(num_classes):
        idx = np.where(np.array(target) == c)[0]
        sc = int(samples_per_class[c])
        assert sc < len(idx), "Not enough labels! len(idx): %d, sample_per_class %d"%(len(idx), sc)
        if determined:
            idx = idx[:sc].tolist()
        else:
            idx = np.random.choice(idx, sc, False).tolist()
        lb_idx = lb_idx + idx
    return lb_idx

def data_subselection(samples, args, flag = None, patch_list = None):
    # generate patch index
    path_list = []
    target_list = []
    for path, target in samples:
        path_list.append(path)
        target_list.append(target)
        
    if args.fixed_order and args.nlabels is not None:
        if os.path.exists(join(args.datadir, args.train_folder, 'order.txt')):
            if args.nlabels < 1:
                num = int(args.nlabels * len(samples))
            else:
                num = args.nlabels
            lb_idx = np.loadtxt(join(args.datadir, args.train_folder, 'order.txt'))[:num].astype(int)
            return list(lb_idx)
    elif patch_list is not None:
        # get patch indices
        patch_idx = []
        for path in path_list:
            # print(path)
            loc = re.search("sq", path).start()
            patch_idx.append(int(path[loc-5:loc]))  
        # we direcly select based on patch index
        patch_uniq_list = np.unique(patch_idx, return_counts = True)[0]
        assert max(patch_list) < len(patch_uniq_list), "wrong patch index"
        # print(patch_uniq_list)
        patch_idx = np.array(patch_idx)
        lb_idx = []
        for p in patch_list:
            # sample per patch and class
            indices = np.where(np.array(patch_idx == patch_uniq_list[p]))[0].tolist()
            if args.half_data: # we want to make sure training and testing are not using the same data
                num = int(len(indices) / 2)
                if flag is not None:
                    if flag == "training":
                        indices_tmp = split_ssl_data([samples[idx][1] for idx in indices], num, determined=True)
                        indices = [indices[idx] for idx in indices_tmp]
                    else:
                        indices_tmp = split_ssl_data([samples[idx][1] for idx in indices], num, determined=True)
                        indices_tmp = [indices[idx] for idx in indices_tmp]
                        indices = list(set(indices) - set(indices_tmp))
            lb_idx = lb_idx + indices
        return lb_idx
    elif args.nlabels is not None:
        return split_ssl_data(target_list, args.nlabels)
    else:
        return list(range(len(samples)))
    

def split_by_patch(samples, args, patch_list = None, flag = None, half_data = False):
    """
    Args:
        samples (list): self.samples from ImageFolder
    """
    # An example for image name "21mar02c_grid1c_00024gr_00017sq_v01_00003hl_v01_00008en-a_crop.png"
    num_labels = args.nlabels
    fixed_order = args.fixed_order
    path_list = []
    target_list = []
    for path, target in samples:
        path_list.append(path)
        target_list.append(target)
    # get patch indices
    patch_idx = []
    for path in path_list:
        # print(path)
        loc = re.search("sq", path).start()
        patch_idx.append(int(path[loc-5:loc]))
    if fixed_order is not None:
        patch_idx = np.array(patch_idx)
        ord = np.argsort(patch_idx)
        N = len(samples)
        m = args.nlabels
        assert m*(fixed_order+1) < N, "Order index too large!"
        lb_idx = np.array(range(N))
        lb_idx = lb_idx[ord][m*fixed_order:m*(fixed_order+1)]
        lb_idx = np.random.choice(lb_idx, num_labels, False)
        return lb_idx.tolist()
    if patch_list is not None:
        patch_uniq_list = np.unique(patch_idx, return_counts = True)[0]
        assert max(patch_list) < len(patch_uniq_list), "wrong patch index"
        # print(patch_uniq_list)
        patch_idx = np.array(patch_idx)
        lb_idx = []
        for p in patch_list:
            # sample per patch and class
            indices = np.where(np.array(patch_idx == patch_uniq_list[p]))[0].tolist()
            if half_data:
                num = int(len(indices) / 2)
                if flag is not None:
                    if flag == "training":
                        indices_tmp = split_ssl_data([samples[idx][1] for idx in indices], num, determined=True)
                        indices = [indices[idx] for idx in indices_tmp]
                    else:
                        indices_tmp = split_ssl_data([samples[idx][1] for idx in indices], num, determined=True)
                        indices_tmp = [indices[idx] for idx in indices_tmp]
                        indices = list(set(indices) - set(indices_tmp))
            lb_idx = lb_idx + indices
        return lb_idx
    
    patch_uniq_list = np.unique(patch_idx, return_counts = True)
    n_path = len(patch_uniq_list[0])
    
    ord = np.random.choice(range(n_path), n_path, False)    
    cum_sum = np.cumsum(patch_uniq_list[1][ord])
    patch_sel_idx = np.where(cum_sum > num_labels)[0][0]
    # print("Kept %d labeled samples" % (cum_sum[patch_sel_idx]))
    keep_patch_list = patch_uniq_list[0][ord][:patch_sel_idx+1]
    
    lb_idx = []
    for p in keep_patch_list:
        # sample per patch and class
        indices = np.where(np.array(patch_idx == p))[0].tolist()
        lb_idx = lb_idx + indices
    lb_idx = np.random.choice(lb_idx, num_labels, False)
    return lb_idx
            
            
            
            
            
    
        
    
        
    
    