from cryoEM_data import CryoEMData
from cryoEM_config import CryoEMConfig
import pickle
import csv
import os
import numpy as np

from tools.data_utils import get_img_names, split_by_name_ts, split_by_name_ctf

CryoEM_DATASETS = {
    'Y3_2Cat':{
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y3_CTF_by_hl.csv',
            'ctf_file': './CryoEM_Config/Y3_CTF_by_hl.csv',
           'category_bins': [0, 6, 999999],
           'feature_dim': 17 
        },
    },
    'Y1_2Cat':{
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
           'category_bins': [0, 6, 999999],
           'feature_dim': 17 
        },
    },
    'Y1_3Cat':{
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y1FULL-3Cat-FM': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_full/FM_resnet18-cosine-bs64-e100/4/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_full/FM_resnet18-cosine-bs64-e100/4/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y1FULL-3Cat-SL': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_full/SL_resnet18-cosine-bs64-e100/1/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_full/SL_resnet18-cosine-bs64-e100/1/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y110-3Cat-FM': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_10/FM_resnet18-cosine-bs16-e100/0/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_10/FM_resnet18-cosine-bs16-e100/0/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y110-3Cat-SL': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_10/SL_resnet18-cosine-bs16-e100/2/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_10/SL_resnet18-cosine-bs16-e100/2/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res50-Y120-3Cat-FM-TrainOnly': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet50-cosine-bs16-e100/1/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet50-cosine-bs16-e100/1/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y110-3Cat-FM-TrainOnly': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet18-cosine-bs16-e100/3/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet18-cosine-bs16-e100/3/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y120-3Cat-FM-TrainOnly': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet18-cosine-bs16-e100/2/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Cat_Train/FM_resnet18-cosine-bs16-e100/2/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y120-3Cat-FM': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_20/FM_resnet18-cosine-bs16-e100/0/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_20/FM_resnet18-cosine-bs16-e100/0/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y120-3Cat-SL': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_20/SL_resnet18-cosine-bs16-e100/0/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu_20/SL_resnet18-cosine-bs16-e100/0/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y15-3Cat-SL': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu/SL_resnet18-cosine-bs16-e200/0/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu/SL_resnet18-cosine-bs16-e200/0/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'Res18-Y15-3Cat-FM': {
        'simple_configure': True,
        'cat_feature': {
            'timestamps': './CryoEM_Config/Y1_timestamps.csv',
            'ctf_file': './CryoEM_Config/Y1_CTF_by_hl.csv',
            'train_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu/FM_resnet18/ur2_pcut_08-cosine-bs16-e50/0/train_by_hl.txt',
           'val_prediction_file': '../CryoEM_Img/log/Y1_3Lab_Quanfu/FM_resnet18/ur2_pcut_08-cosine-bs16-e50/0/val_by_hl.txt',
           'category_bins': [0, 4, 8, 999999],
           'feature_dim': 18 
        },
    },
    'CryoEM-resnet18': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_categorization_res18_train_by_hl.txt',
           'val_prediction_file': './CryoEM_data/2_categorization_res18_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet18-train-cnn-feature.pickle',
           'val_visual_file': './CryoEM_data/resnet18-val-cnn-feature.pickle',
           'category_bins': [0, 6, 999999],
           'feature_dim': 17
        },
    },
    'CryoEM-resnet50': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
           'train_prediction_file': './CryoEM_data/2_categorization_res50_train_by_hl.txt',
           'val_prediction_file': './CryoEM_data/2_categorization_res50_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet50-train-cnn-feature.pickle',
           'val_visual_file': './CryoEM_data/resnet50-val-cnn-feature.pickle',
           'category_bins': [0, 6, 999999],
           'feature_dim': 17
        },
    },
    'CryoEM-weighted-resnet50': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_categorization_weighted_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_categorization_weighted_res50_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet50-train-cnn-feature.pickle',
            'val_visual_file': './CryoEM_data/resnet50-val-cnn-feature.pickle',
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        },
    },
    'CryoEM-8bit-resnet18': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_cat_8bit_res18_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_cat_8bit_res18_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet18-train-cnn-feature-8bit.pickle',
            'val_visual_file': './CryoEM_data/resnet18-val-cnn-feature-8bit.pickle',
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        }
    },
    'CryoEM-8bit-ctf4-resnet18': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_cat_8bit_ctf4_res18_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_cat_8bit_ctf4_res18_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet18-train-cnn-feature-8bit-ctf4.pickle',
            'val_visual_file': './CryoEM_data/resnet18-val-cnn-feature-8bit-ctf4.pickle',
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        }
    },
    'CryoEM-8bit-resnet50': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_cat_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_cat_8bit_res50_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet50-train-cnn-feature-8bit.pickle',
            'val_visual_file': './CryoEM_data/resnet50-val-cnn-feature-8bit.pickle',
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        },
        'regress_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-ctf4-resnet50': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_cat_8bit_ctf4_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_cat_8bit_ctf4_res50_val_by_hl.txt',
            'train_visual_file': './CryoEM_data/resnet50-train-cnn-feature-8bit-ctf4.pickle',
            'val_visual_file': './CryoEM_data/resnet50-val-cnn-feature-8bit-ctf4.pickle',
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        },
        'regress_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/2_regress_8bit_ctf4_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/2_regress_8bit_ctf4_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-resnet50-M': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/CTF_train_by_hl.csv',
            'val_ctf_file': './CryoEM_data/CTF_val_by_hl.csv',
            'train_prediction_file': './CryoEM_data/Y1_M_2_cat_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y1_M_2_cat_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        },
    },
    'CryoEM-8bit-resnet50-M-Y2': {
        'cat_feature': {
            'train_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'val_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'train_ctf_file': './CryoEM_data/Y2_M_2_cat_8bit_res50_train.txt',
            'val_ctf_file': './CryoEM_data/Y2_M_2_cat_8bit_res50_val.txt',
            'train_prediction_file': None,
            'val_prediction_file': None,
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 6, 999999],
            'feature_dim': 17
        },
    },
    'CryoEM-8bit-resnet50-Y1': {
        'regress_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/timestamps.csv',
            'train_ctf_file': './CryoEM_data/target_CTF_Y1_train.csv',
            'val_ctf_file': './CryoEM_data/target_CTF_Y1_val.csv',
            'train_prediction_file': './CryoEM_data/Y1_2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y1_2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-resnet50-Y2': {
        'regress_feature': {
            'train_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'val_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'train_ctf_file': './CryoEM_data/target_CTF_Y2_train.csv',
            'val_ctf_file': './CryoEM_data/target_CTF_Y2_val.csv',
            'train_prediction_file': './CryoEM_data/Y2Y2_2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y2Y2_2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-resnet50-Y2-new': {
        'regress_feature': {
            'train_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'val_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'train_ctf_file': './CryoEM_data/target_CTF_Y2_train.csv',
            'val_ctf_file': './CryoEM_data/target_CTF_Y2_val.csv',
            'train_prediction_file': './CryoEM_data/Y2_2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y2_2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-resnet50-MFT2-Y2-new': {
        'regress_feature': {
            'train_timestamps': './CryoEM_data/timestamps.csv',
            'val_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'train_ctf_file': './CryoEM_data/target_CTF_Y1_train.csv',
            'val_ctf_file': './CryoEM_data/target_CTF_Y2_val.csv',
            'train_prediction_file': './CryoEM_data/Y1_2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y2FT_2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
    'CryoEM-8bit-resnet50-M-Y3-new': {
        'regress_feature': {
            'train_timestamps': './CryoEM_data/target_CTF_Y2.csv',
            'val_timestamps': './CryoEM_data/target_CTF_Y3.csv',
            'train_ctf_file': './CryoEM_data/target_CTF_Y2_train.csv',
            'val_ctf_file': './CryoEM_data/target_CTF_Y3_val.csv',
            'train_prediction_file': './CryoEM_data/Y2_2_regress_8bit_res50_train_by_hl.txt',
            'val_prediction_file': './CryoEM_data/Y3_2_regress_8bit_res50_val_by_hl.txt',
            'train_visual_file': None,
            'val_visual_file': None,
            'category_bins': [0, 4, 6, 8, 10, 999999],
            'feature_dim': 32
        }
    },
}

def get_cnn_features(filename, image_list):
    with open(filename , 'rb') as f:
        features = pickle.load(f)

    output = np.stack([features[item] for item in image_list])
    print ('----', output.shape)
    return output

def get_dataset(dataset, prediction_type, use_one_hot=False, visual_similarity=False, prediction_file=None):
    dataset = CryoEM_DATASETS[dataset]

    feature_set = dataset['cat_feature'] if prediction_type == CryoEMConfig.CLASSIFICATION else dataset['regress_feature']
    if prediction_file is not None:
        
        feature_set['train_prediction_file'] = os.path.join('../CryoEM_Img', prediction_file, 'train_by_hl.txt')
        feature_set['val_prediction_file'] = os.path.join('../CryoEM_Img', prediction_file, 'val_by_hl.txt')
    feature_dim = feature_set['feature_dim']
    category_bins = feature_set['category_bins']

    # set a mode
    if dataset['simple_configure'] is not None:
        # find train and val index and then split timestamp and ctf files
        # with open(feature_set['timestamps'], 'r') as csvfile:
        #     csvreader = csv.reader(csvfile)
        #     timestamps = [(row[0],row[1]) for row in csvreader]
        # with open(feature_set['ctf_file'], 'r') as csvfile:
        #     csvreader = csv.reader(csvfile)
        #     ctfs = [(row[0],row[3], row[4]) for row in csvreader]
        train_idx = get_img_names(feature_set['train_prediction_file'])
        val_idx = get_img_names(feature_set['val_prediction_file'])
        
        # let us remove training from validation
        # val_num = np.array(list(filter(lambda x: val_idx[x] not in train_idx, range(len(val_idx)))))
        
        # val_file = np.loadtxt(feature_set['val_prediction_file'], delimiter=' ', dtype=str)[val_num, :]
        val_file = np.loadtxt(feature_set['val_prediction_file'], delimiter=' ', dtype=str)
        ctfs_train = split_by_name_ctf(feature_set['ctf_file'], train_idx)
        ctfs_val = split_by_name_ctf(feature_set['ctf_file'], val_idx)
        train_dataset = CryoEMData(feature_set['timestamps'],
                                ctf_file=ctfs_train,
                                prediction_file=feature_set['train_prediction_file'],
                                prediction_type=prediction_type,
                                category_bins=category_bins,
                                use_one_hot=use_one_hot)

        val_dataset = CryoEMData(feature_set['timestamps'],
                       ctf_file=ctfs_val,
                       prediction_file=val_file,
                       prediction_type=prediction_type,
                       category_bins=category_bins,
                       use_one_hot=use_one_hot) 
        
    else:
        train_dataset = CryoEMData(feature_set['train_timestamps'],
                                ctf_file=feature_set['train_ctf_file'],
                                prediction_file=feature_set['train_prediction_file'],
                                prediction_type=prediction_type,
                                category_bins=category_bins,
                                use_one_hot=use_one_hot)

        val_dataset = CryoEMData(feature_set['val_timestamps'],
                       ctf_file=feature_set['val_ctf_file'],
                       prediction_file=feature_set['val_prediction_file'],
                       prediction_type=prediction_type,
                       category_bins=category_bins,
                       use_one_hot=use_one_hot)

    train_visual_feature = None
    val_visual_feature = None
    if visual_similarity:
        train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
        train_visual_feature = get_cnn_features(feature_set['train_visual_file'], train_file_list)
        val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
        val_visual_feature = get_cnn_features(feature_set['val_visual_file'], val_file_list)
    print(train_dataset.num_holes())
    print(val_dataset.num_holes())
    return train_dataset, val_dataset, feature_dim, category_bins, train_visual_feature, val_visual_feature


'''
def get_dataset(dataset, prediction_type, use_one_hot=False, visual_similarity=False, gaussian_filter_size=0.0):
    train_visual_feature = None
    val_visual_feature = None
    category_bins = [0, 6, 999999]
    feature_dim = 17
    elif dataset == 'CryoEM-resnet18':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_categorization_res18_train_by_hl.txt',
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_categorization_res18_val_by_hl.txt',
                                 use_one_hot=use_one_hot)

        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet18-train-cnn-feature.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet18-val-cnn-feature.pickle', val_file_list)

    elif dataset == 'CryoEM-resnet50':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_categorization_res50_train_by_hl.txt',
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_categorization_res50_val_by_hl.txt',
                                 use_one_hot=use_one_hot)
        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet50-train-cnn-feature.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet50-val-cnn-feature.pickle', val_file_list)
    elif dataset == 'CryoEM-weighted-resnet50':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_categorization_weighted_res50_train_by_hl.txt',
                                    use_one_hot=use_one_hot)
        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_categorization_weighted_res50_val_by_hl.txt',
                                 use_one_hot=use_one_hot)
    elif dataset == 'CryoEM-8bit-resnet18':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_res18_train_by_hl.txt',
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_res18_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)

        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet18-train-cnn-feature-8bit.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet18-val-cnn-feature-8bit.pickle', val_file_list)

    elif dataset == 'CryoEM-8bit-resnet50':
        if prediction_type == CryoEMConfig.CLASSIFICATION:
            feature_dim = 17
            category_bins = [0, 6, 999999]
            train_file = './CryoEM_data/2_cat_8bit_res50_train_by_hl.txt'
            val_file =  './CryoEM_data/2_cat_8bit_res50_val_by_hl.txt'
        else:
            feature_dim = 32
            category_bins = [0, 4, 6, 8, 10, 999999]
            train_file = './CryoEM_data/2_regress_8bit_res50_train_by_hl.txt'
            val_file =  './CryoEM_data/2_regress_8bit_res50_val_by_hl.txt'

        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file=train_file,
                                    prediction_type=prediction_type,
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file=val_file,
                                 prediction_type=prediction_type,
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet50-train-cnn-feature-8bit.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet50-val-cnn-feature-8bit.pickle', val_file_list)
    elif dataset == 'CryoEM-8bit-weighted-resnet18':
        feature_dim = 17
        category_bins = [0, 6, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_weighted_res18_train_by_hl.txt',
                                    category_bins=category_bins,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_weighted_res18_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
    elif dataset == 'CryoEM-8bit-weighted-resnet50':
        feature_dim = 17
        category_bins = [0, 6, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_weighted_res50_train_by_hl.txt',
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_weighted_res50_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
    elif dataset == 'CryoEM-8bit-ctf4-resnet18':
        feature_dim = 17
        category_bins = [0, 4, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_ctf4_res18_train_by_hl.txt',
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_ctf4_res18_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)

        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet18-train-cnn-feature-8bit-ctf4.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet18-val-cnn-feature-8bit-ctf4.pickle', val_file_list)

    elif dataset == 'CryoEM-8bit-ctf4-resnet50':
        feature_dim = 17
        category_bins = [0, 4, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_ctf4_res50_train_by_hl.txt',
                                     category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_ctf4_res50_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet50-train-cnn-feature-8bit-ctf4.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet50-val-cnn-feature-8bit-ctf4.pickle', val_file_list)
    elif dataset == 'CryoEM-8bit-ctf4-weighted-resnet18':
        feature_dim = 17
        category_bins = [0, 4, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_ctf4_weighted_res18_train_by_hl.txt',
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_ctf4_weighted_res18_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
    elif dataset == 'CryoEM-8bit-ctf4-weighted-resnet50':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/2_cat_8bit_ctf4_weighted_res50_train_by_hl.txt',
                                    category_bins=category_bins,
                                    gaussian_filter_size = gaussian_filter_size,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/2_cat_8bit_ctf4_weighted_res50_val_by_hl.txt',
                                 category_bins=category_bins,
                                 gaussian_filter_size=gaussian_filter_size,
                                 use_one_hot=use_one_hot)
    elif dataset == 'CryoEM-8bit-ctf4-cat3-resnet18':
        feature_dim = 21
        category_bins = [0, 4, 8, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/3_cat_8bit_ctf4_res18_train_by_hl.txt',
                                    category_bins=category_bins,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/3_cat_8bit_ctf4_res18_val_by_hl.txt',
                                 category_bins=category_bins,
                                 use_one_hot=use_one_hot)

        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet18-train-cnn-feature-8bit-ctf4-cat3.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet18-val-cnn-feature-8bit-ctf4-cat3.pickle', val_file_list)

    elif dataset == 'CryoEM-8bit-ctf4-cat3-resnet50':
        feature_dim = 21
        category_bins = [0, 4, 8, 999999]
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train_by_hl.csv',
                                    prediction_file='./CryoEM_data/3_cat_8bit_ctf4_res50_train_by_hl.txt',
                                     category_bins=category_bins,
                                    use_one_hot=use_one_hot)

        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val_by_hl.csv',
                                 prediction_file='./CryoEM_data/3_cat_8bit_ctf4_res50_val_by_hl.txt',
                                 category_bins=category_bins,
                                 use_one_hot=use_one_hot)
        if visual_similarity:
            train_file_list = [train_dataset.get_hole(k).name for k in range(train_dataset.num_holes())]
            train_visual_feature = get_cnn_features('./CryoEM_data/resnet50-train-cnn-feature-8bit-ctf4.pickle', train_file_list)
            val_file_list = [val_dataset.get_hole(k).name for k in range(val_dataset.num_holes())]
            val_visual_feature = get_cnn_features('./CryoEM_data/resnet50-val-cnn-feature-8bit-ctf4.pickle', val_file_list)
    elif dataset == 'CryoEM-7-3-Category':
        train_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                    ctf_file='./CryoEM_data/CTF_train.csv',
                                    prediction_file='./CryoEM_data/2_categorization_train.txt',
                                    use_one_hot=use_one_hot)
        val_dataset = CryoEMData('./CryoEM_data/timestamps.csv',
                                 ctf_file='./CryoEM_data/CTF_val.csv',
                                 prediction_file='./CryoEM_data/2_categorization_val.txt',
                                 use_one_hot=use_one_hot)
    else:
        raise ValueError('Dataset {} not available.'.format(dataset))

    return train_dataset, val_dataset, feature_dim, category_bins, train_visual_feature, val_visual_feature
'''
