import os
import gym
import torch
import pickle
import pprint
import argparse
import numpy as np
import cProfile
import pstats
import warnings
warnings.filterwarnings("ignore")

from cryoEM_dataset import get_dataset
from cryoEM_env import CryoEMEnv
from cryoEM_config import *
#from dqn import Net
from pathlib import Path
import copy
import time
import pandas as pd
import scipy.stats as stats


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', type=str, default='Y3_2Cat')
    parser.add_argument('--seed', type=int, default=1626)
    parser.add_argument('--ctf-thresh', type=float, default=6.0)
    parser.add_argument('--prediction_file', type=str, default="/gpfs/wscgpfs02/zipingxu/CryoEM_Img/log/timeBased/0721111708/")
    args = parser.parse_args()
    print(args)
    return args

def get_dataframe(dataset):
    num_holes = dataset.num_holes()
    ctf_dict = []
    for k in range(num_holes):
        g_idx, s_idx, p_idx, h_idx = dataset.idx(k)
        ctf =  dataset.get_hole(k)._gt_ctf.value
        res = []
        s = ''
        for to_app in [g_idx, s_idx, p_idx, h_idx]:
            s += str(to_app)
            res.append(s)
        res.append(ctf)
        ctf_dict.append(res)
    dat = pd.DataFrame(ctf_dict, columns=['grid', 'square', 'patch', 'hole', 'ctf'])
    return dat


def get_anova(sample_df, by_col = 'grid', value = 'ctf'):
    # Create ANOVA backbone table
    data = [['Between Groups', '', '', '', '', '', ''], ['Within Groups', '', '', '', '', '', ''], ['Total', '', '', '', '', '', '']] 
    anova_table = pd.DataFrame(data, columns = ['Source of Variation', 'SS', 'df', 'MS', 'F', 'P-value', 'F crit']) 
    anova_table.set_index('Source of Variation', inplace = True)

    # calculate SSTR and update anova table
    x_bar = sample_df[value].mean()
    SSTR = sample_df.groupby(by_col).count() * (sample_df.groupby(by_col).mean() - x_bar)**2
    anova_table['SS']['Between Groups'] = SSTR[value].sum()

    # calculate SSE and update anova table
    SSE = (sample_df.groupby(by_col).count() - 1) * sample_df.groupby(by_col).std()**2
    anova_table['SS']['Within Groups'] = SSE[value].sum()

    # calculate SSTR and update anova table
    SSTR = SSTR[value].sum() + SSE[value].sum()
    anova_table['SS']['Total'] = SSTR

    # update degree of freedom
    anova_table['df']['Between Groups'] = sample_df[by_col].nunique() - 1
    anova_table['df']['Within Groups'] = sample_df.shape[0] - sample_df[by_col].nunique()
    anova_table['df']['Total'] = sample_df.shape[0] - 1

    # calculate MS
    anova_table['MS'] = anova_table['SS'] / anova_table['df']

    # calculate F 
    F = anova_table['MS']['Between Groups'] / anova_table['MS']['Within Groups']
    anova_table['F']['Between Groups'] = F

    # p-value
    anova_table['P-value']['Between Groups'] = 1 - stats.f.cdf(F, anova_table['df']['Between Groups'], anova_table['df']['Within Groups'])

    # F critical 
    alpha = 0.05
    # possible types "right-tailed, left-tailed, two-tailed"
    tail_hypothesis_type = "two-tailed"
    if tail_hypothesis_type == "two-tailed":
        alpha /= 2
    anova_table['F crit']['Between Groups'] = stats.f.ppf(1-alpha, anova_table['df']['Between Groups'], anova_table['df']['Within Groups'])

    # Final ANOVA Table
    return anova_table


args = get_args()
train_dataset, val_dataset, _, _, _, _ = get_dataset(args.dataset, prediction_type=CryoEMConfig.CLASSIFICATION, use_one_hot=False, prediction_file=args.prediction_file)
dat = get_dataframe(train_dataset)

grid = get_anova(dat, by_col = 'grid')
square = get_anova(dat, by_col = 'square')
patch = get_anova(dat, by_col = 'patch')
hole = get_anova(dat, by_col = 'hole')